from __future__ import annotations

import dataclasses
from typing import Optional

from ..executors import Executor, PlainTextExecutorResult


@dataclasses.dataclass(frozen=True)
class RemoteSpecification:
    """A model specification from a remote URL"""

    url: str
    """The specification's http(s) URL"""

    async def fetch_sources(self, executor: Executor) -> list[str]:
        async with executor.execute(
            result_type=PlainTextExecutorResult,
            url=self.url,
        ) as res:
            source = await res.text(assert_status=200)
        return [source]


@dataclasses.dataclass(frozen=True)
class FormulationSpecification:
    """A specification from an `Optimization Hub`_ formulation

    This type of specification allows queueing solves and is recommended for
    production use as it provides history and reproducibility when combined
    with tag names.

    .. _Optimization Hub: https://hub.cloud.opvious.io
    """

    formulation_name: str
    """The corresponding formulation's name"""

    tag_name: Optional[str] = None
    """The matching tag's name

    If absent, the formulation's default will be used.
    """



import logging

from .client import Client, ClientSetting, Problem
from .common import __version__
from .data.outcomes import (
    AbortedOutcome,
    FailedOutcome,
    FeasibleOutcome,
    InfeasibleOutcome,
    SolveOutcome,
    SolveStatus,
    UnboundedOutcome,
    UnexpectedSolveOutcomeError,
    solve_outcome_status,
)
from .data.outlines import Label, ProblemOutline
from .data.queued_solves import QueuedSolve, SolveNotification
from .data.solves import (
    EpsilonConstraint,
    Solution,
    SolveInputs,
    SolveOptions,
    SolveOutputs,
    ProblemSummary,
    SolveStrategy,
    Target,
)
from .data.tensors import (
    DimensionArgument,
    Key,
    KeyItem,
    SparseTensorArgument,
    Tensor,
    TensorArgument,
    Value,
)
from . import executors
from . import modeling
from .specifications import (
    FormulationSpecification,
    LocalSpecification,
    LocalSpecificationIssue,
    LocalSpecificationSource,
    RemoteSpecification,
    Specification,
    load_notebook_models,
)
from . import transformations


__all__ = [
    # Client
    "Client",
    "ClientSetting",
    # Executors
    "executors",
    # Specifications
    "FormulationSpecification",
    "LocalSpecification",
    "LocalSpecificationIssue",
    "LocalSpecificationSource",
    "RemoteSpecification",
    "Specification",
    "load_notebook_models",
    "modeling",
    # Solves
    "DimensionArgument",
    "EpsilonConstraint",
    "Key",
    "KeyItem",
    "Label",
    "Problem",
    "ProblemOutline",
    "ProblemSummary",
    "QueuedSolve",
    "Solution",
    "SolveInputs",
    "SolveNotification",
    "SolveOptions",
    "SolveOutputs",
    "SolveStatus",
    "SolveStrategy",
    "SparseTensorArgument",
    "Target",
    "Tensor",
    "TensorArgument",
    "Value",
    # Outcomes
    "AbortedOutcome",
    "FailedOutcome",
    "FeasibleOutcome",
    "InfeasibleOutcome",
    "SolveOutcome",
    "UnboundedOutcome",
    "UnexpectedSolveOutcomeError",
    "solve_outcome_status",
    # Transformations
    "transformations",
    # Miscellaneous
    "__version__",
]


logging.getLogger(__name__).addHandler(logging.NullHandler())



import contextlib
import logging
from pyodide.http import pyfetch  # type: ignore
from typing import AsyncIterator, Optional

from .common import (
    CONTENT_TYPE_HEADER,
    Executor,
    ExecutorResult,
    JsonExecutorResult,
    Headers,
    PlainTextExecutorResult,
    TRACE_HEADER,
    unsupported_content_type_error,
)


_logger = logging.getLogger(__name__)


class PyodideExecutor(Executor):
    """`pyodide`-powered executor"""

    def __init__(self, endpoint: str, authorization: Optional[str] = None):
        super().__init__(
            variant="pyodide", endpoint=endpoint, authorization=authorization
        )

    @contextlib.asynccontextmanager
    async def _send(
        self, url: str, method: str, headers: Headers, body: Optional[bytes]
    ) -> AsyncIterator[ExecutorResult]:
        res = await pyfetch(
            url=url,
            method=method,
            headers=headers,
            body=body,
        )
        status = res.status
        headers = res.js_response.headers
        trace = headers.get(TRACE_HEADER)
        ctype = headers.get(CONTENT_TYPE_HEADER)
        if JsonExecutorResult.is_eligible(ctype):
            text = await res.js_response.text()
            yield JsonExecutorResult(status=status, trace=trace, text=text)
        elif PlainTextExecutorResult.is_eligible(ctype):
            text = await res.js_response.text()
            yield PlainTextExecutorResult(
                status=status, trace=trace, reader=text
            )
        else:
            text = await res.js_response.text()
            raise unsupported_content_type_error(
                text=text,
                content_type=ctype,
                trace=trace,
            )



"""Built-in model fragments

This module exports :class:`~opvious.modeling.ModelFragment` instances for
common use-cases.
"""

from __future__ import annotations

from typing import Any, Callable, Iterable, Optional, Union

from ..common import method_decorator, untuple
from .ast import cross, domain, lift, Projection, Quantifiable, total
from .definitions import (
    Constraint,
    Expression,
    ExpressionLike,
    Image,
    Parameter,
    Tensor,
    TensorLike,
    Variable,
    alias,
    constraint,
)
from .identifiers import Name
from .model import ModelFragment
from .quantified import Quantified


class MaskedSubspace(ModelFragment):
    """Masked subspace fragment"""

    def __init__(
        self,
        *quantifiables: Quantifiable,
        alias_name: Optional[Name] = None,
    ) -> None:
        self._alias_name = alias_name
        self._quantifiables = quantifiables
        self._mask = Parameter.indicator(quantifiables)

    @property
    def mask(self) -> Parameter:
        """Parameter controlling the subset's element"""
        return self._mask

    @property
    @alias(lambda init, self: init(self._alias_name))
    def masked(self) -> Quantified:
        """Masked subset

        As a convenience, iterating on the subset directly also yields
        quantifiers from the masked subset.
        """
        for t in cross(self._quantifiables):
            if self._mask(*t):
                yield untuple(t)

    def __iter__(self) -> Iterable[Any]:
        return (untuple(t) for t in cross(self.masked))


MaskedSubset = MaskedSubspace  # Deprecated alias


class DerivedVariable(ModelFragment):
    """Variable equal to a given equation

    Args:
        body: The equation defining the variable's value
        quantifiables: Variable quantification
        name: Name of the generated variable
        image: Generated variable :class:`~opvious.modeling.Image`
    """

    default_definition = "value"

    def __init__(
        self,
        body: Callable[..., Any],
        *quantifiables: Quantifiable,
        name: Optional[Name] = None,
        image: Image = Image(),
    ) -> None:
        self._body = body
        self._value = Variable(image, quantifiables, name=name)

    @property
    def value(self) -> Variable:
        return self._value

    @constraint
    def is_defined(self) -> Quantified:
        for t in self._value.space():
            yield self._value(*t) == self._body(*t)

    def __call__(self, *subs: ExpressionLike) -> Expression:
        return self._value(*subs)


@method_decorator(require_call=True)
def derived_variable(
    *quantifiables: Quantifiable,
    name: Optional[Name] = None,
    image: Image = Image(),
) -> Callable[[TensorLike], DerivedVariable]:
    """Transforms a method into a :class:`DerivedVariable` fragment"""

    def wrapper(fn):
        return DerivedVariable(fn, quantifiables, name=name, image=image)

    return wrapper


class MagnitudeVariable(ModelFragment):
    """Absolute value variable fragment

    Args:
        tensor: Non-negative tensor-like
        quantifiables: Tensor quantifiables. Can be omitted if `tensor` is a
            variable or parameter.
        image: Tensor image. Defaults to `tensor`'s if it is a variable or
            parameter, else non-negative reals.
        name: Name of the generated magnitude variable
        projection: Mask used to project the variable's quantification
        lower_bound: Disable the lower bound
        upper_bound: Disable the upper bound

    See also :func:`magnitude_variable` for a decorator equivalent.
    """

    default_definition = "value"

    def __init__(
        self,
        tensor: TensorLike,
        *quantifiables: Quantifiable,
        name: Optional[Name] = None,
        image: Optional[Image] = None,
        projection: Projection = -1,
        lower_bound=True,
        upper_bound=True,
    ) -> None:
        if isinstance(tensor, Tensor):
            if not quantifiables:
                quantifiables = tensor.quantifiables()
            if not image:
                image = tensor.image
            if lower_bound and tensor.image.lower_bound == 0:
                lower_bound = False
            if upper_bound and tensor.image.upper_bound == 0:
                upper_bound = False
        self._tensor = tensor
        self._domains = tuple(domain(q) for q in quantifiables)
        self._projection = projection
        self._lower_bound = lower_bound
        self._upper_bound = upper_bound
        self._value = Variable(
            image or Image(lower_bound=0),
            self._quantification(),
            name=name,
        )

    def _quantification(self, lift=False):
        return cross(*self._domains, projection=self._projection, lift=lift)

    @property
    def value(self) -> Variable:
        """The magnitude variable"""
        return self._value

    @constraint(lambda init, self: init(disabled=not self._lower_bound))
    def lower_bounds(self) -> Quantified:
        """The magnitude's lower bound constraint"""
        for cp in self._quantification(lift=True):
            yield -self.value(*cp) <= self._tensor(*cp.lifted)

    @constraint(lambda init, self: init(disabled=not self._upper_bound))
    def upper_bounds(self) -> Quantified:
        """The magnitude's upper bound constraint"""
        for cp in self._quantification(lift=True):
            yield self._value(*cp) >= self._tensor(*cp.lifted)

    def __call__(self, *subs: ExpressionLike) -> Expression:
        return self._value(*subs)


@method_decorator()
def magnitude_variable(
    *quantifiables: Quantifiable,
    name: Optional[Name] = None,
    image: Optional[Image] = None,
    projection: Projection = -1,
    lower_bound=True,
    upper_bound=True,
) -> Callable[[Callable[..., TensorLike]], MagnitudeVariable]:
    """Transforms a method into a :class:`MagnitudeVariable` fragment

    Note that this method may alter the underlying method's call signature if a
    projection is specified.
    """

    def wrapper(fn):
        return MagnitudeVariable(
            fn,
            *quantifiables,
            name=name,
            image=image,
            projection=projection,
            lower_bound=lower_bound,
            upper_bound=upper_bound,
        )

    return wrapper


Magnitude = MagnitudeVariable  # Deprecated alias


class ActivationVariable(ModelFragment):
    """Indicator variable activation fragment

    This variable tracks an underlying tensor or tensor-like expression.
    Assuming both of its bounds are defined (see below) it will be equal to 1
    iff the underlying tensor is positive and 0 otherwise.

    Args:
        tensor: Tensor-like
        quantifiables: Tensor quantifiables, can be omitted if the tensor is a
            variable or parameter
        upper_bound: Value of the upper bound used in the activation
            constraint. If `True` the variable's image's upper bound will be
            used, if `False` no activation constraint will be added.
        lower_bound: Value of the lower bound used in the deactivation
            constraint. If `True` the variable's image's lower bound will be
            used, if `False` no deactivation constraint will be added.
        name: Name of the generated activation variable
        negate: Negate the returned indicator variable.
        projection: Mask used to project the variable's quantification. When
            this is set, the indicator variable will be set to 1 iff at least
            one of the projected tensor values is positive.
    """

    # This fragment shows an alternate implementation which does not rely on
    # lazy method decorators. It uses a closure instead of instance variables.
    def __new__(
        cls,
        tensor: TensorLike,
        *quantifiables: Quantifiable,
        upper_bound: Union[ExpressionLike, TensorLike, bool] = True,
        lower_bound: Union[ExpressionLike, TensorLike, bool] = False,
        name: Optional[Name] = None,
        negate=False,
        projection: Projection = -1,
    ) -> ActivationVariable:
        if not quantifiables and isinstance(tensor, Tensor):
            quantifiables = tensor.quantifiables()
        domains = tuple(domain(q) for q in quantifiables)

        def quantification(lift=False, projection=projection):
            return cross(*domains, projection=projection, lift=lift)

        def tensor_image():
            if not isinstance(tensor, Tensor):
                raise ValueError(
                    f"Cannot infer bound for tensor-like {tensor}"
                )
            return tensor.image

        class _Fragment(ActivationVariable):
            value = Variable.indicator(quantification(), name=name)

            def __new__(cls) -> _Fragment:
                return ModelFragment.__new__(cls)

            def __call__(self, *subs: ExpressionLike) -> Expression:
                return self.value(*subs)

            @constraint(disabled=upper_bound is False)
            def activates(self):
                bound = upper_bound
                for cp in quantification(lift=True):
                    if callable(bound):
                        bound = bound(*cp.lifted)
                    elif bound is True:
                        bound = tensor_image().upper_bound
                    value = 1 - self.value(*cp) if negate else self.value(*cp)
                    yield bound * value >= tensor(*cp.lifted)

            @constraint(disabled=lower_bound is False)
            def deactivates(self):
                bound = lower_bound
                for cp in quantification():
                    if projection >= 0:
                        term = total(
                            tensor(*lift(cp, ep, projection))
                            for ep in quantification(projection=~projection)
                        )
                    else:
                        term = tensor(*cp)
                    if callable(bound):
                        bound = bound(*cp)
                    elif bound is True:
                        bound = tensor_image().lower_bound
                    value = 1 - self.value(*cp) if negate else self.value(*cp)
                    yield bound * value <= term

        return _Fragment()

    default_definition = "value"

    @property
    def value(self) -> Variable:
        """Activation variable value

        As a convenience calling the fragment directly also returns expressions
        from this variable.
        """
        raise NotImplementedError()

    @property
    def activates(self) -> Optional[Constraint]:
        """Constraint ensuring that the activation variable activates

        This constraint enforces that the activation variable is set to 1 when
        at least one the underlying tensor's value is positive.
        """
        raise NotImplementedError()

    @property
    def deactivates(self) -> Optional[Constraint]:
        """Constraint ensuring that the activation variable deactivates

        This constraint enforces that the activation variable is set to 0 when
        none of the underlying tensor's values are non-zero. It requires the
        fragment to have a non-zero lower bound.
        """
        raise NotImplementedError()

    def __call__(self, *subs: ExpressionLike) -> Expression:
        raise NotImplementedError()


ActivationIndicator = ActivationVariable  # Deprecated alias


@method_decorator(require_call=True)
def activation_variable(
    *quantifiables: Quantifiable,
    upper_bound: Union[ExpressionLike, TensorLike, bool] = True,
    lower_bound: Union[ExpressionLike, TensorLike, bool] = False,
    name: Optional[Name] = None,
    negate=False,
    projection: Projection = -1,
) -> Callable[[Callable[..., TensorLike]], ActivationVariable]:
    """Transforms a method into an :class:`ActivationVariable` fragment

    Note that this method may alter the underlying method's call signature if a
    projection is specified.
    """

    def wrapper(fn):
        return ActivationVariable(
            fn,
            *quantifiables,
            lower_bound=lower_bound,
            upper_bound=upper_bound,
            name=name,
            negate=negate,
            projection=projection,
        )

    return wrapper



from __future__ import annotations

import collections
import dataclasses
import glob
import logging
import os
from typing import Iterable, Mapping, Optional, Sequence

from ..common import Json


_DEFAULT_TITLE = "untitled"


_logger = logging.getLogger(__name__)


@dataclasses.dataclass(frozen=True)
class LocalSpecificationSource:
    """A document containing one or more model definitions

    Models can comprise multiple sources, with each source only containing a
    subset of the model's definitions. This provides an easy way to share
    definitions between models.
    """

    text: str
    """The source's contents"""

    title: str = _DEFAULT_TITLE
    """A title used to easily identify the source"""

    def _repr_markdown_(self) -> str:
        return _source_details(self, [], True)


@dataclasses.dataclass(frozen=True)
class LocalSpecificationAnnotation:
    """API-generated annotation"""

    issue_count: int
    """The total number of issues detected in this specification"""

    issues: Mapping[int, Sequence[LocalSpecificationIssue]]
    """All issues associated with the specification, keyed by source index"""


@dataclasses.dataclass(frozen=True)
class LocalSpecificationIssue:
    """An issue detected within a specification"""

    source_index: int
    start_offset: int
    end_offset: int
    message: str
    code: str


def local_specification_issue_from_json(data: Json) -> LocalSpecificationIssue:
    rg = data["range"]
    return LocalSpecificationIssue(
        message=data["message"],
        source_index=data["index"],
        start_offset=rg["start"]["offset"],
        end_offset=rg["end"]["offset"],
        code=data["code"],
    )


_SPECIFICATION_STYLE = " ".join(
    [
        "margin-top: 1em;",
        "margin-bottom: 1em;",
    ]
)


_SUMMARY_STYLE = " ".join(
    [
        "cursor: pointer;",
        "text-decoration: underline;",
        "text-decoration-style: dotted;",
    ]
)


@dataclasses.dataclass(frozen=True)
class LocalSpecification:
    """
    A local model specification

    Instances of this class are integrated with IPython's `rich display
    capabilities`_ and will automatically render their LaTeX sources when
    output in notebooks.

    Note that this type of specification cannot be used to queue solves.

    .. _rich display capabilities: https://ipython.readthedocs.io/en/stable/config/integrating.html#rich-display  # noqa
    """

    sources: Sequence[LocalSpecificationSource]
    """The model's mathematical source definitions"""

    description: Optional[str] = None
    """Optional description"""

    annotation: Optional[LocalSpecificationAnnotation] = None
    """API-issued annotation

    This field is generated automatically by clients'
    :meth:`~opvious.Client.annotate_specification` method. When any issues are
    detected, the specification's pretty-printed representation will highlight
    any errors.
    """

    def source(self, title: Optional[str] = None) -> LocalSpecificationSource:
        """Returns the first source, optionally matching a given title"""
        if title is None:
            return self.sources[0]
        for source in self.sources:
            if source.title == title:
                return source
        raise Exception(f"Missing source for title {title}")

    @classmethod
    def inline(cls, *texts: str) -> LocalSpecification:
        """Creates a local specification from inline LaTeX definitions"""
        sources = [LocalSpecificationSource(s) for s in texts]
        return LocalSpecification(sources)

    @classmethod
    def globs(
        cls, *likes: str, root: Optional[str] = None
    ) -> LocalSpecification:
        """Creates a local specification from LaTeX definition files

        Args:
            *likes: File names or globs
            root: Optional root folder used for matching file names and globs.
                As a convenience this argument can also be an existing file's
                name, in which case it will be interpreted as its parent
                directory (this is typically handy when used with `__file__`).
        """
        if root:
            root = os.path.realpath(root)
            if os.path.isfile(root):
                root = os.path.dirname(root)
        sources: list[LocalSpecificationSource] = []
        for like in likes:
            for path in glob.iglob(like, root_dir=root, recursive=True):
                if root:
                    path = os.path.join(root, path)
                with open(path, "r", encoding="utf-8") as reader:
                    sources.append(
                        LocalSpecificationSource(
                            text=reader.read(),
                            title=path,
                        )
                    )
        return LocalSpecification(sources)

    def annotated(
        self, issues: Iterable[LocalSpecificationIssue]
    ) -> LocalSpecification:
        count = 0
        grouped: dict[
            int, list[LocalSpecificationIssue]
        ] = collections.defaultdict(list)
        for issue in issues:
            count += 1
            grouped[issue.source_index].append(issue)
        annotation = LocalSpecificationAnnotation(
            issues=dict(grouped),
            issue_count=count,
        )
        return dataclasses.replace(self, annotation=annotation)

    def _repr_markdown_(self) -> str:
        if self.annotation:
            issues = self.annotation.issues
            for index, group in issues.items():
                messages = [f"\t* {i.message} [{i.code}]" for i in group]
                _logger.error(
                    "%s issue(s) in specification '%s':\n%s",
                    len(group),
                    self.sources[index].title,
                    "\n".join(messages),
                )
        else:
            issues = {}
        contents = "\n---\n".join(
            _source_details(s, issues.get(i) or [], i == 0)
            for i, s in enumerate(self.sources)
        )
        return f'<div style="{_SPECIFICATION_STYLE}">{contents}</div>'


def _source_details(
    source: LocalSpecificationSource,
    issues: Sequence[LocalSpecificationIssue],
    start_open=False,
) -> str:
    summary = source.title
    if issues:
        summary += " &#9888;"
    return "\n".join(
        [
            "",
            "<details open>" if start_open else "<details>",
            f'<summary style="{_SUMMARY_STYLE}">{summary}</summary>',
            '<div style="margin-top: 1em;">',
            _colorize(source.text, issues),
            "</div>",
            "</details>",
            "",
        ]
    )


def _colorize(text: str, issues: Sequence[LocalSpecificationIssue]) -> str:
    if not issues:
        return text

    by_start = sorted(issues, key=lambda i: i.start_offset)
    cutoffs: list[int] = []
    end = 0
    for issue in by_start:
        if issue.code == "ERR_INVALID_SYNTAX":
            # These issues can't be easily colorized as they span name
            # boundaries
            continue
        if issue.start_offset <= end:
            end = max(issue.end_offset + 1, end)
        else:
            cutoffs.append(end)
            cutoffs.append(issue.start_offset)
            end = issue.end_offset + 1
    cutoffs.append(end)
    cutoffs.append(len(text))

    ret: list[str] = []
    for i, piece in enumerate(text[m:n] for m, n in zip(cutoffs, cutoffs[1:])):
        is_error = i % 2 == 1
        if is_error:
            ret.append(f"\\color{{orange}}{{{piece}}}")
        else:
            ret.append(piece)
    return "".join(ret)



import base64
import io
import pandas as pd
import sys
from typing import Optional

from .common import (
    Executor,
    ExecutorError,
    ExecutorResult,
    JsonExecutorResult,
    JsonSeqExecutorResult,
    PlainTextExecutorResult,
)


__all__ = [
    "authorization_header",
    "aiohttp_executor",
    "default_executor",
    "pyodide_executor",
    "urllib_executor",
    "default_executor",
    "Executor",
    "ExecutorError",
    "ExecutorResult",
    "JsonExecutorResult",
    "JsonSeqExecutorResult",
    "PlainTextExecutorResult",
    "fetch_csv",
    "fetch_text",
]


def aiohttp_executor(
    endpoint: str, authorization: Optional[str] = None
) -> Executor:
    """Returns an `aiohttp`-powered executor"""
    from .aiohttp import AiohttpExecutor

    return AiohttpExecutor(endpoint, authorization)


def pyodide_executor(
    endpoint: str, authorization: Optional[str] = None
) -> Executor:
    """Returns a Pyodide-compatible executor"""
    from .pyodide import PyodideExecutor

    return PyodideExecutor(endpoint=endpoint, authorization=authorization)


def urllib_executor(
    endpoint: str, authorization: Optional[str] = None
) -> Executor:
    """Returns a native `urllib` executor"""
    from .urllib import UrllibExecutor

    return UrllibExecutor(endpoint=endpoint, authorization=authorization)


def _is_using_pyodide():
    # https://pyodide.org/en/stable/usage/faq.html#how-to-detect-that-code-is-run-with-pyodide
    return "pyodide" in sys.modules


def default_executor(
    endpoint: str, authorization: Optional[str] = None
) -> Executor:
    """Infers the best executor for the current environment"""
    if _is_using_pyodide():
        return pyodide_executor(endpoint=endpoint, authorization=authorization)
    try:
        return aiohttp_executor(endpoint=endpoint, authorization=authorization)
    except ImportError:
        return urllib_executor(endpoint=endpoint, authorization=authorization)


def authorization_header(token: str) -> str:
    """Generates a suitable authorization header from a token"""
    if " " in token:
        return token
    if ":" in token:
        value = base64.b64encode(token.encode("utf8")).decode("utf8")
        return f"Basic {value}"
    return f"Bearer {token}"


async def fetch_text(url: str) -> str:
    """Fetches a URL's contents as plain text

    Args:
        url: The URL to fetch (via GET)
    """
    # We specify an invalid endpoint URL so that default headers are never
    # included in the request. Otherwise it causes certain requests to fail
    # (e.g. when fetching gists).
    executor = default_executor("ignored://")
    async with executor.execute(PlainTextExecutorResult, url) as res:
        return await res.text()


async def fetch_csv(url: str, **kwargs) -> pd.DataFrame:
    """Fetches a CSV from a URL

    This convenience method is provided as a portable way to fetch data from
    notebooks. Using the default `pandas.read_csv` does not support `https://`
    URLs from JupyterLite (Pyodide).

    Args:
        url: The URL to fetch (via GET)
        **kwargs: Keyword arguments forwarded to `pandas.read_csv`
    """
    text = await fetch_text(url)
    return pd.read_csv(io.StringIO(text), **kwargs)



from __future__ import annotations

import collections
import contextlib
import contextvars
import dataclasses
from typing import (
    Any,
    Generator,
    Iterable,
    KeysView,
    Mapping,
    Optional,
    Sequence,
)

from ..common import Label


Name = str


Environment = KeysView[Name]


@dataclasses.dataclass(eq=False, frozen=True)
class Identifier:
    def format(self) -> Name:
        """Generates the final name"""
        return _active_scope.get().format(self)


class GlobalIdentifier(Identifier):
    name: Optional[Name]


@dataclasses.dataclass(eq=False, frozen=True)
class DimensionIdentifier(GlobalIdentifier):
    name: Optional[Name]


@dataclasses.dataclass(eq=False, frozen=True)
class TensorIdentifier(GlobalIdentifier):
    name: Optional[Name]
    is_parameter: bool


@dataclasses.dataclass(eq=False, frozen=True)
class AliasIdentifier(GlobalIdentifier):
    name: Name


@dataclasses.dataclass(eq=False, frozen=True)
class QuantifierGroup:
    alias: AliasIdentifier
    subscripts: tuple[Any, ...]
    rank: int


class QuantifierIdentifier(Identifier):
    space: Any  # ScalarSpace
    groups: Sequence[QuantifierGroup]
    name: Optional[Name]

    @classmethod
    def base(cls, space: Any) -> QuantifierIdentifier:
        return _BaseQuantifierIdentifier(space=space, groups=[])

    def grouped_within(self, group: QuantifierGroup) -> QuantifierIdentifier:
        return _BaseQuantifierIdentifier(
            space=self.space, groups=[group, *self.groups]
        ).named(self.name)

    def named(self, name: Optional[Name]) -> QuantifierIdentifier:
        if name is None or name == self.name:
            return self
        return _NamedQuantifierIdentifier(name=name, parent=self)

    @property
    def outer_group(self) -> Optional[QuantifierGroup]:
        return self.groups[0] if self.groups else None


@dataclasses.dataclass(eq=False, frozen=True)
class _BaseQuantifierIdentifier(QuantifierIdentifier):
    space: Any
    groups: Sequence[QuantifierGroup]
    name = None


@dataclasses.dataclass(eq=False, frozen=True)
class _NamedQuantifierIdentifier(QuantifierIdentifier):
    name: Name
    parent: QuantifierIdentifier

    @property
    def space(self) -> Any:
        return self.parent.space

    @property
    def groups(self) -> Sequence[QuantifierGroup]:  # type: ignore[override]
        return self.parent.groups


class IdentifierFormatter:
    def __init__(self, labels: Mapping[GlobalIdentifier, Label]) -> None:
        self._formatted: dict[GlobalIdentifier, bool] = {}
        self._labels = labels

    def formatted_globals(self) -> Sequence[GlobalIdentifier]:
        return list(self._formatted)

    def format(self, identifier: Identifier, env: Environment) -> Name:
        if isinstance(identifier, GlobalIdentifier):
            self._formatted[identifier] = True
            if identifier.name:
                return identifier.name
            try:
                label = self._labels[identifier]
            except KeyError as exc:
                raise Exception(
                    "Unknown identifier, did you forget to add a dependency?"
                ) from exc
            if isinstance(identifier, DimensionIdentifier):
                return self._format_dimension(label, env)
            elif isinstance(identifier, TensorIdentifier):
                if identifier.is_parameter:
                    return self._format_parameter(label, env)
                else:
                    return self._format_variable(label, env)
        elif isinstance(identifier, QuantifierIdentifier):
            return self.format_quantifier(identifier, env)
        raise TypeError(f"Unexpected identifier: {identifier}")

    def _format_dimension(self, label: Label, env: Environment) -> Name:
        raise NotImplementedError()

    def _format_parameter(self, label: Label, env: Environment) -> Name:
        raise NotImplementedError()

    def _format_variable(self, label: Label, env: Environment) -> Name:
        raise NotImplementedError()

    def format_quantifier(
        self, identifier: QuantifierIdentifier, env: Environment
    ) -> Name:
        raise NotImplementedError()


@dataclasses.dataclass
class _Scope:
    formatter: IdentifierFormatter
    environment: collections.ChainMap[Name, Identifier]  # To owning identifier
    quantifier_names: collections.ChainMap[QuantifierIdentifier, Name]
    global_names: dict[GlobalIdentifier, Name]

    def child(self) -> _Scope:
        return dataclasses.replace(
            self,
            environment=self.environment.new_child(),
            quantifier_names=self.quantifier_names.new_child(),
        )

    def format(self, identifier: Identifier) -> Name:
        if isinstance(identifier, GlobalIdentifier):
            name = self.global_names.get(identifier)
        else:
            if not isinstance(identifier, QuantifierIdentifier):
                raise TypeError(f"Unexpected identifier: {identifier}")
            name = self.quantifier_names.get(identifier)
            if name is None:
                raise Exception(f"Missing quantifier: {identifier}")
        if name is None:
            name = self.prepare(identifier)
        return name

    def prepare(self, identifier: Identifier) -> Name:
        name = self.formatter.format(identifier, self.environment.keys())
        owner = self.environment.get(name)
        if owner is not None and owner != identifier:
            raise Exception(f"Name collision: {name}")
        if isinstance(identifier, GlobalIdentifier):
            self.global_names[identifier] = name
            self.environment.maps[-1][name] = identifier
        else:
            if not isinstance(identifier, QuantifierIdentifier):
                raise TypeError(f"Unexpected identifier: {identifier}")
            self.environment[name] = identifier
            self.quantifier_names[identifier] = name
            while isinstance(identifier, _NamedQuantifierIdentifier):
                identifier = identifier.parent
                self.quantifier_names[identifier] = name
        return name


_active_scope: Any = contextvars.ContextVar("formatting_scope")


@contextlib.contextmanager
def global_formatting_scope(
    formatter: IdentifierFormatter,
    reserved: Optional[Mapping[Name, GlobalIdentifier]] = None,
) -> Generator[None, None, None]:
    scope = _active_scope.get(None)
    if scope:
        raise Exception("Identifier formatter already active")
    scope = _Scope(
        formatter=formatter,
        environment=collections.ChainMap(dict(reserved) if reserved else {}),
        quantifier_names=collections.ChainMap(),
        global_names={},
    )
    token = _active_scope.set(scope)
    try:
        yield
    finally:
        _active_scope.reset(token)


@contextlib.contextmanager
def local_formatting_scope(
    quantifiers: Iterable[QuantifierIdentifier],
) -> Generator[None, None, None]:
    scope = _active_scope.get(None)
    if not scope:
        raise Exception("Missing active formatter")
    child = scope.child()
    for quantifier in quantifiers:
        child.prepare(quantifier)
    token = _active_scope.set(child)
    try:
        yield
    finally:
        _active_scope.reset(token)


class DefaultIdentifierFormatter(IdentifierFormatter):
    def __init__(self, labels: Mapping[GlobalIdentifier, Label]) -> None:
        super().__init__(labels)

    def _format_dimension(self, label: Label, env: Environment) -> Name:
        i = _last_capital_index(label)
        r = label[i or 0].upper()
        n = r if i is None else f"{r}^\\mathrm{{{label[:i]}}}"
        return _first_available(n, env)

    def _format_parameter(self, label: Label, env: Environment) -> Name:
        i = _last_capital_index(label)
        r = label[i or 0].lower()
        n = r if i is None else f"{r}^\\mathrm{{{label[:i]}}}"
        return _first_available(n, env)

    def _format_variable(self, label: Label, env: Environment) -> Name:
        i = _last_capital_index(label)
        r = label[i or 0].lower()
        g = _greek_letters.get(r, r)
        n = g if i is None else f"{g}^\\mathrm{{{label[:i]}}}"
        return _first_available(n, env)

    def format_quantifier(
        self, identifier: QuantifierIdentifier, env: Environment
    ) -> Name:
        name = identifier.name
        if not name:
            sp = identifier.space
            group = None
            for g in identifier.groups:
                if g.rank == 1:
                    # Single rank alias
                    group = g
                    break
            if not group and hasattr(sp, "identifier") and sp.identifier:
                # Dimension
                name = _lower_principal(sp.identifier.format())
            else:
                # Interval, possibly aliased
                if not group:
                    group = identifier.outer_group
                if group:
                    name = _lower_principal(group.alias.format())
        return _first_available(name or _DEFAULT_QUANTIFIER_NAME, env)


_DEFAULT_QUANTIFIER_NAME = "x"


def _first_available(name: Name, env: Environment) -> Name:
    while name in env:
        if name.endswith("}"):
            name = name[:-1] + r" \prime}"
        else:
            name += "'"
    return name


def _last_capital_index(label: Label) -> Optional[int]:
    j = None
    for i, c in enumerate(label):
        if c.isupper():
            j = i
    return j


def _lower_principal(name: Name) -> Name:
    if "^" not in name:
        return name.lower()
    parts = name.split("^", 1)
    return f"{parts[0].lower()}^{parts[1]}"


_greek_letters = {
    "a": "\\alpha",
    "b": "\\beta",
    "c": "\\chi",
    "d": "\\delta",
    "e": "\\epsilon",
    "f": "\\phi",
    "g": "\\gamma",
    "h": "\\eta",
    "i": "\\iota",
    "j": "\\xi",  # TODO: Find better alternative
    "k": "\\kappa",
    "l": "\\lambda",
    "m": "\\mu",
    "n": "\\nu",
    "o": "\\omicron",
    "p": "\\pi",
    "q": "\\theta",
    "r": "\\rho",
    "s": "\\sigma",
    "t": "\\tau",
    "u": "\\psi",
    "v": "\\zeta",  # TODO: Find better alternative
    "w": "\\omega",
    "x": "\\xi",
    "y": "\\upsilon",
    "z": "\\zeta",
}



import asyncio
import dataclasses
import docopt
import json
import os.path
import sys
from typing import Any, Mapping, Optional

from .common import __version__
from .client import Client
from .specifications import load_notebook_models, LocalSpecification

_COMMAND = "python -m opvious"

_DOC = f"""Opvious Python CLI

Command line utilities for interacting with the Opvious API from Python. The
underlying client can be configured using the `OPVIOUS_ENDPOINT` and
`OPVIOUS_TOKEN` environment variables. Refer to the standard CLI
(https://www.npmjs.com/package/opvious-cli) for additional operations.

Usage:
    {_COMMAND} register-notebook PATH [MODEL]
        [-dn NAME] [-t TAGS] [--allow-empty]
    {_COMMAND} register-sources GLOB [-dn NAME] [-t TAGS]
    {_COMMAND} (-h | --help)
    {_COMMAND} --version

Options:
    --allow-empty       Do not throw an error if no models were found in a
                        notebook
    -d, --dry-run       Validate the specification but do not store it on the
                        server
    -n, --name NAME     Formulation name. By default this name is inferred
                        from the file's name, omitting the extension
    -t, --tags TAGS     Comma-separated list of tags. By default only the
                        `latest` tag is added
    --version           Show SDK version
    -h, --help          Show this message
"""


class _SpecificationHandler:
    def __init__(
        self, client: Client, tags: Optional[str] = None, dry_run=False
    ) -> None:
        self._client = client
        self._tags = tags.split() if tags else None
        self._dry_run = dry_run

    async def _handle(self, spec: LocalSpecification, name: str) -> None:
        if self._dry_run:
            annotated = await self._client.annotate_specification(spec)
            ann = annotated.annotation
            assert ann is not None
            if ann.issue_count > 0:
                raise Exception(
                    f"{ann.issue_count} issue(s) in specification: "
                    + json.dumps(
                        [
                            dataclasses.asdict(i)
                            for issues in ann.issues.values()
                            for i in issues
                        ]
                    )
                )
            print("Specification is valid")
        else:
            await self._client.register_specification(
                spec,
                formulation_name=name,
                tag_names=self._tags,
            )
            print(f"Registered specification in formulation {name}")

    async def handle_notebook(
        self,
        path: str,
        model_name: Optional[str],
        name: Optional[str],
        allow_empty: bool,
    ) -> None:
        sn = load_notebook_models(path, allow_empty=allow_empty)
        if model_name is None:
            model_names = list(sn.__dict__.keys())
            if not self._dry_run and len(model_names) != 1:
                raise Exception(f"Notebook has 0 or 2+ models ({model_names})")
        else:
            model_names = [model_name]
        if name is None:
            name = _default_name(path)
        for model_name in model_names:
            model = getattr(sn, model_name)
            await self._handle(model.specification(), name)

    async def handle_sources(self, glob: str, name: Optional[str]) -> None:
        if name is None:
            name = _default_name(glob)
        spec = LocalSpecification.globs(glob)
        await self._handle(spec, name)


def _default_name(path: str) -> str:
    return os.path.splitext(os.path.basename(path))[0]


async def _run(args: Mapping[str, Any]) -> None:
    handler = _SpecificationHandler(
        Client.default(), args["--tags"], args["--dry-run"]
    )
    if args["register-notebook"]:
        await handler.handle_notebook(
            args["PATH"],
            model_name=args["MODEL"],
            name=args["--name"],
            allow_empty=args["--allow-empty"],
        )
    elif args["register-sources"]:
        await handler.handle_sources(args["GLOB"], name=args["--name"])
    else:
        raise Exception("Unexpected command")


if __name__ == "__main__":
    argv = _COMMAND.split()[1:] + sys.argv[1:]
    args = docopt.docopt(_DOC, argv=argv, version=__version__)
    asyncio.run(_run(args))



import functools
from importlib import metadata
import math
from typing import Any, Callable, Iterable, Union
import urllib.parse
import weakref


try:
    __version__ = metadata.version(__package__)
except metadata.PackageNotFoundError:
    __version__ = ""


del metadata


# Formatting


Label = str


def format_percent(val):
    if val == "Infinity":
        return "inf"
    return f"{int(val * 10_000) / 100}%"


def is_url(s: str) -> bool:
    """Checks if a string is a URL."""
    try:
        res = urllib.parse.urlparse(s)
        return bool(res.scheme and res.netloc)
    except ValueError:
        return False


def to_camel_case(s: str) -> str:
    if "_" not in s:
        return s
    return "".join(
        p.capitalize() if i else p for i, p in enumerate(s.split("_")) if p
    )


def untuple(t: Iterable[Any]) -> Any:
    if not isinstance(t, tuple):
        t = tuple(t)
    return t[0] if len(t) == 1 else t


# JSON utilities


Json = Any


ExtendedFloat = Union[float, str]


def encode_extended_float(val: ExtendedFloat) -> Json:
    if val == math.inf:
        return "Infinity"
    elif val == -math.inf:
        return "-Infinity"
    return val


def decode_extended_float(val: ExtendedFloat) -> Json:
    if val == "Infinity":
        return math.inf
    elif val == "-Infinity":
        return -math.inf
    return val


def json_dict(**kwargs) -> Json:
    """Strips keys with None values and encodes infinity values"""
    data = {}
    for key, val in kwargs.items():
        if val is None:
            continue
        json_key = to_camel_case(key)
        if isinstance(val, float):
            data[json_key] = encode_extended_float(val)
        else:
            data[json_key] = val
    return data


# Decorator utilities


_lambda = lambda: 0  # noqa


def _is_lambda(fn: Callable[..., Any]) -> bool:
    return fn.__name__ == _lambda.__name__


def capturing_instance(wrapper: Callable[..., Any]) -> Any:
    def wrap(fn):
        return Bindable(fn, wrapper)

    return wrap


def with_instance(consumer: Callable[..., Any]) -> Any:
    def wrap(fn):
        return Bindable(fn, consumer, lazy=True)

    return wrap


def method_decorator(require_call=False):
    """Transforms a decorator into a method-friendly equivalent"""

    def wrap_decorator(decorator: Callable[..., Any]) -> Any:
        def wrapped_decorator(*args, **kwargs):
            arg = args[0] if args else None
            if callable(arg):

                if _is_lambda(arg):
                    # Lazy decorator constructor
                    if len(args) > 1 or kwargs:
                        raise Exception("Unexpected tail arguments")

                    def wrap_method(meth):
                        return Bindable(
                            meth, lambda self: arg(decorator, self), lazy=True
                        )

                    return wrap_method
                elif not require_call and len(args) == 1 and not kwargs:
                    # No argument decorator
                    return Bindable(arg, decorator())

            # Standard decorator creation

            def wrap_method(meth):
                return Bindable(meth, decorator(*args, **kwargs))

            return wrap_method

        return wrapped_decorator

    return wrap_decorator


class Bindable:
    def __init__(
        self, body: Callable[..., Any], wrapper: Callable[..., Any], lazy=False
    ) -> None:
        self._body = body
        self._wrapper = wrapper
        self._lazy = lazy
        self._bindings: Any = weakref.WeakKeyDictionary()

    def _apply(self, owner: Any, bind=True) -> Any:
        wrapper = self._wrapper(owner) if self._lazy else self._wrapper
        body = functools.partial(self._body, owner) if bind else self._body
        return wrapper(body)

    def bound_to(self, owner: Any) -> Any:
        binding = self._bindings.get(owner)
        if not binding:
            binding = self._apply(owner)
            while isinstance(binding, Bindable):
                binding = binding._apply(owner, False)
            self._bindings[owner] = binding
        return binding

    def __get__(self, owner: Any, _objtype=None) -> Any:
        return self.bound_to(owner)

    def __call__(self, owner, *args, **kwargs) -> Any:
        # Needed for property calls and direct calls
        return self.bound_to(owner)(*args, **kwargs)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['opvious',
 'opvious.client',
 'opvious.data',
 'opvious.executors',
 'opvious.modeling',
 'opvious.specifications']

package_data = \
{'': ['*']}

install_requires = \
['backoff>=2.2,<3.0',
 'humanize>=4.4.0,<5.0.0',
 'importnb>=2023.1.7,<2024.0.0',
 'numpy>=1.21',
 'pandas>=1.4']

extras_require = \
{'aio': ['aiohttp>=3.8,<4.0', 'Brotli>=1.0.9,<2.0.0'],
 'cli': ['docopt>=0.6.2,<0.7.0']}

setup_kwargs = {
    'name': 'opvious',
    'version': '0.18.6rc1',
    'description': 'Opvious Python SDK',
    'long_description': '# Opvious Python SDK  [![CI](https://github.com/opvious/sdk.py/actions/workflows/ci.yml/badge.svg)](https://github.com/opvious/sdk.py/actions/workflows/ci.yml) [![Pypi badge](https://badge.fury.io/py/opvious.svg)](https://pypi.python.org/pypi/opvious/)\n\n<div align="center">\n  <p>\n    Define and deploy optimization models in minutes with <a href="https://www.opvious.io">Opvious</a>.\n  </p>\n  <a href="https://www.opvious.io/notebooks/retro/notebooks/?path=guides/welcome.ipynb"><img src="https://www.opvious.io/opvious-steps.png" style="height: 600px;"/></a>\n    <p>\n    <a href="https://www.opvious.io/notebooks/retro/notebooks/?path=guides/welcome.ipynb">Try it out!</a>\n  </p>\n</div>\n\n\n## Highlights\n\n+ Declarative modeling API with __extensive static checks and automatic LaTeX generation__\n+ Remote solves with __real-time progress notifications__\n+ __Flexible data import/export__ via `pandas`\n+ __Advanced multi-objective support__: weighted sums, epsilon constraints, ...\n+ __Smart debugging capabilities__: semantic constraint relaxations, annotated LP formatting, ...\n\n\n## Documentation\n\n+ [Getting started guide](https://www.opvious.io/notebooks/retro/notebooks/?path=guides/welcome.ipynb)\n+ [SDK API reference](https://opvious.readthedocs.io)\n+ [Interactive optimization notebooks](https://github.com/opvious/notebooks)\n\n\n## Licensing\n\nThe SDK is licensed under Apache 2.0. The Opvious platform is available both as a managed cloud service (free on small datasets) and self-hosted (free for non-commercial projects). See our [plans page](https://www.opvious.io/plans) for more information.\n',
    'author': 'Opvious Engineering',
    'author_email': 'oss@opvious.io',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/opvious/sdk.py',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.9,<3.12',
}


setup(**setup_kwargs)



"""Built-in transformations

This module exports all available :class:`~opvious.ProblemTransformation`
instances.
"""

from __future__ import annotations

import dataclasses
import math
from typing import Literal

from .common import Json, json_dict
from .data.outlines import Label, ProblemOutline
from .data.solves import Target


class ProblemTransformationContext:
    def __init__(self):
        self._json = []

    def add(self, kind: str, **kwargs) -> None:
        self.add_json(json_dict(kind=kind, **kwargs))

    def add_json(self, data: Json) -> None:
        self._json.append(data)

    def get_json(self) -> list[Json]:
        return self._json

    async def fetch_outline(self) -> ProblemOutline:
        raise NotImplementedError()


class ProblemTransformation:
    """Base problem transformation class

    You should not need to interact with this class directly, instead use one
    of the available :ref:`transformation subclasses <Base transformations>`.
    """

    async def register(self, context: ProblemTransformationContext) -> None:
        raise NotImplementedError()


@dataclasses.dataclass(frozen=True)
class PinVariables(ProblemTransformation):
    """A transformation which pins one or more variables to input values

    Each pinned variable will have an associated derived parameter labeled
    `$label_pin` with the same domain and image as the original variable. For
    example a non-negative `production` variable would have a non-negative
    parameter labeled `production_pin`.

    A constraint (labeled `$label_isPinned`) is also automatically added which
    enforces equality between the variable and any inputs passed in to the
    parameter. Values for keys which do not have a pin remain free, so it is
    possible to do partial pinning. For example, assuming a monthly variable
    labeled `production`, the following code would only pin production in
    January (to 100):

    .. code-block:: python

        transformations = [opvious.transformations.PinVariables("production")]
        parameters = {
            "production_pin": {"january": 100},
            # ...
        }

    Finally, note that this constraint can also be transformed similar to any
    other constraint. For example it can sometimes be useful to relax it via
    :class:`.RelaxConstraints`.
    """

    labels: list[Label] = dataclasses.field(default_factory=lambda: [])
    """The labels of the variables to pin

    If empty, all variables will be pinned.
    """

    async def register(self, context: ProblemTransformationContext) -> None:
        labels = self.labels
        if not labels:
            outline = await context.fetch_outline()
            labels = list(outline.variables)
        for label in labels:
            context.add("pinVariable", label=label)


RelaxationPenalty = Literal[
    "TOTAL_DEVIATION",
    "MAX_DEVIATION",
    "DEVIATION_CARDINALITY",
]


@dataclasses.dataclass(frozen=True)
class RelaxConstraints(ProblemTransformation):
    """A transformation which relaxes one or more constraints

    Each relaxed constraint will be omitted from the formulation and replaced
    by a slack variable and an objective minimizing this slack (two of each for
    equality constraints). The derived variables have the same domain as the
    relaxed constraint, are always non-negative, and are labeled as follows:

    + `$label_deficit` for deficit slack (applicable for greater than and
      equality constraints).
    + `$label_surplus` for surplus slack (applicable for less than and equality
      constraints).

    For example an equality constraint labeled `isBalanced` would be
    transformed into two variables labeled `isBalanced_deficit` and
    `isBalanced_surplus`. A greater than constraint labeled `demandIsMet` would
    be relaxed into a single variable labeled `demandIsMet_deficit`.

    Each deficit (resp. surplus) variable has a corresponding objective labeled
    `$label_minimizeDeficit` (resp. `$label_minimizeSurplus`). Refer to the
    `penalty` parameter for details on how slack is penalized.

    Finally, since relaxed formulations will almost always have multiple
    objectives, you may also need to specific a :class:`opvious.SolveStrategy`.
    A common pattern is to omit any existing objectives and minimize the
    aggregate slack violation (see :ref:`Detecting infeasibilities`).
    """

    labels: list[Label] = dataclasses.field(default_factory=lambda: [])
    """The labels of the constraints to relax

    If empty, all constraints will be relaxed.
    """

    penalty: RelaxationPenalty = "TOTAL_DEVIATION"
    """Slack penalization mode

    + `TOTAL_DEVIATION`: Cost proportional to the total sum of the (absolute
      value of) slack for the constraint.
    + `MAX_DEVIATION`: Cost proportional to the maximum deviation for the
      constraint.
    + `DEVIATION_CARDINALITY`: Cost proportional to the number of rows with
      non-zero deviation. This penalty requires the relaxation to be capped.

    The default is `TOTAL_DEVIATION`.
    """

    is_capped: bool = False
    """Whether slack is capped

    Setting this to true will create an additional parameter for each slack
    variable labeled `$label_deficitCap` (resp. `$label_surplusCap`) for
    deficit (resp. surplus). This parameter has the same domain as the variable
    and is non-negative.

    This option is required when using the `DEVIATION_CARDINALITY` penalty.
    """

    async def register(self, context: ProblemTransformationContext) -> None:
        labels = self.labels
        if not labels:
            outline = await context.fetch_outline()
            labels = list(outline.constraints)
        for label in labels:
            context.add(
                "relaxConstraint",
                label=label,
                penalty=self.penalty,
                is_capped=self.is_capped,
            )


@dataclasses.dataclass(frozen=True)
class DensifyVariables(ProblemTransformation):
    """A transformation which updates one or more variables to be continuous"""

    labels: list[Label] = dataclasses.field(default_factory=lambda: [])
    """The labels of the variables to densify

    If empty, all integral variables will be densified.
    """

    async def register(self, _context: ProblemTransformationContext) -> None:
        raise NotImplementedError()  # TODO


@dataclasses.dataclass(frozen=True)
class OmitConstraints(ProblemTransformation):
    """A transformation which drops one or more constraints

    Any parameters or variables which are not referenced in any remaining
    constraint or objective will automatically be dropped.
    """

    labels: list[Label] = dataclasses.field(default_factory=lambda: [])
    """The labels of the constraints to drop

    If empty, all constraints will be dropped.
    """

    async def register(self, context: ProblemTransformationContext) -> None:
        labels = self.labels
        if not labels:
            outline = await context.fetch_outline()
            labels = list(outline.constraints)
        for label in labels:
            context.add("omitConstraint", label=label)


@dataclasses.dataclass(frozen=True)
class OmitObjectives(ProblemTransformation):
    """A transformation which drops one or more objectives

    Similar to :class:`.OmitConstraints`, any parameters or variables which are
    not referenced in any remaining constraint or objective will automatically
    be dropped.
    """

    labels: list[Label] = dataclasses.field(default_factory=lambda: [])
    """The labels of the objectives to drop

    If empty, all objectives will be dropped.
    """

    async def register(self, context: ProblemTransformationContext) -> None:
        labels = self.labels
        if not labels:
            outline = await context.fetch_outline()
            labels = list(outline.objectives)
        for label in labels:
            context.add("omitObjective", label=label)


@dataclasses.dataclass(frozen=True)
class ConstrainObjective(ProblemTransformation):
    """A transformation which bounds the value of a objective

    This can be used for example to guarantee a minimum objective levels when
    multiple objectives are involved or to implement custom multi-objective
    strategies (see :ref:`Weighted distance multi-objective optimization`).
    """

    target: Target
    """The label of the objective to constrain"""

    min_value: float = -math.inf
    """The objective's minimum allowed value"""

    max_value: float = math.inf
    """The objective's maximum allowed value"""

    async def register(self, _context: ProblemTransformationContext) -> None:
        raise NotImplementedError()  # TODO



from __future__ import annotations

import collections
import dataclasses
import itertools
import math
from typing import (
    Any,
    Iterable,
    Iterator,
    Mapping,
    Optional,
    Protocol,
    Sequence,
    TypeVar,
    Union,
    cast,
    overload,
)

from ..common import untuple
from .identifiers import (
    AliasIdentifier,
    Identifier,
    Name,
    local_formatting_scope,
    QuantifierGroup,
    QuantifierIdentifier,
)
from .quantified import Quantified, unquantify, declare


def render_identifier(iden: Identifier, *subscripts: Expression) -> str:
    s = iden.format()
    if subscripts:
        sub = ",".join(s.render() for s in subscripts)
        s += f"_{{{sub}}}"
    return s


# https://docs.python.org/3/reference/datamodel.html#emulating-numeric-types
class Expression:
    """Base expression

    Expressions are typically created via :class:`.Parameter` or
    :class:`.Variable` instances. They can be combined using any of the
    following operations:

    * Addition: `x + y`
    * Substraction: `x - y`
    * Multiplication: `x * y`
    * Modulo: `x % y`
    * Integer division: `x // y`
    * Floating division: `x / y`
    * Power: `x ** y`

    Literal numbers will automatically be promoted to expressions when used in
    any of the operations above. They may also be manually wrapped with
    :func:`literal`.

    Other types of expressions may be created by one of the functions below:

    * :func:`total` for summations
    * :func:`cross` for quantifiers
    * :func:`size` for cardinality expressions
    * :func:`switch` for branching logic

    See also :class:`.Predicate` for the list of supported comparison
    operators.
    """

    def __neg__(self) -> Expression:
        return _BinaryExpression("mul", literal(-1), self)

    def __abs__(self) -> Expression:
        return _UnaryExpression("abs", self)

    def __add__(self, other: ExpressionLike) -> Expression:
        if is_literal(self, 0):
            return to_expression(other)
        if is_literal(other, 0):
            return self
        return _BinaryExpression("add", self, to_expression(other))

    def __radd__(self, left: ExpressionLike) -> Expression:
        return to_expression(left) + self

    def __sub__(self, other: ExpressionLike) -> Expression:
        if is_literal(self, 0):
            return -to_expression(other)
        if is_literal(other, 0):
            return self
        return _BinaryExpression("sub", self, to_expression(other))

    def __rsub__(self, left: ExpressionLike) -> Expression:
        return to_expression(left) - self

    def __mod__(self, other: ExpressionLike) -> Expression:
        return _BinaryExpression("mod", self, to_expression(other))

    def __rmod__(self, left: ExpressionLike) -> Expression:
        return to_expression(left) % self

    def __mul__(self, other: ExpressionLike) -> Expression:
        if is_literal(self, 1):
            return to_expression(other)
        if is_literal(other, 1):
            return self
        return _BinaryExpression("mul", self, to_expression(other))

    def __rmul__(self, left: ExpressionLike) -> Expression:
        return to_expression(left) * self

    def __truediv__(self, other: ExpressionLike) -> Expression:
        return _BinaryExpression("div", self, to_expression(other))

    def __rtruediv__(self, left: ExpressionLike) -> Expression:
        return to_expression(left) / self

    def __floordiv__(self, other: ExpressionLike) -> Expression:
        inner = _BinaryExpression("div", self, to_expression(other))
        return _UnaryExpression("floor", inner)

    def __rfloordiv__(self, left: ExpressionLike) -> Expression:
        return to_expression(left) // self

    def __pow__(self, other: ExpressionLike) -> Expression:
        return _BinaryExpression("pow", self, to_expression(other))

    def __rpow__(self, left: ExpressionLike) -> Expression:
        return to_expression(left) ** self

    def __lt__(self, other: ExpressionLike) -> Predicate:
        return _ComparisonPredicate("<", self, to_expression(other))

    def __le__(self, other: ExpressionLike) -> Predicate:
        return _ComparisonPredicate("\\leq", self, to_expression(other))

    def __eq__(self, other: object) -> Predicate:  # type: ignore[override]
        if not isinstance(other, (Expression, float, int)):
            return NotImplemented
        return _ComparisonPredicate("=", self, to_expression(other))

    def __ne__(self, other: object) -> Predicate:  # type: ignore[override]
        if not isinstance(other, (Expression, float, int)):
            return NotImplemented
        return _ComparisonPredicate("\\neq", self, to_expression(other))

    def __gt__(self, other: ExpressionLike) -> Predicate:
        return _ComparisonPredicate(">", self, to_expression(other))

    def __ge__(self, other: ExpressionLike) -> Predicate:
        return _ComparisonPredicate("\\geq", self, to_expression(other))

    def __bool__(self) -> bool:
        return bool(self != 0)

    def render(self, _precedence=0) -> str:
        raise NotImplementedError()


ExpressionLike = Union[Expression, float, int]


@dataclasses.dataclass(eq=False, frozen=True)
class _LiteralExpression(Expression):
    value: float

    def render(self, _precedence=0) -> str:
        if self.value == math.inf:
            return "\\infty"
        elif self.value == -math.inf:
            return "-\\infty"
        else:
            return str(self.value)


def literal(val: Union[float, int]) -> Expression:
    """Wraps a literal value into an expression

    Arg:
        val: float or integer

    In general you will not need to use this method as expression operators
    automatically call this under the hood.
    """
    if not isinstance(val, (float, int)):
        raise TypeError("Unexpected literal value")
    return _LiteralExpression(val)


def is_literal(expr: ExpressionLike, val: Union[float, int]) -> bool:
    if not isinstance(expr, Expression):
        return expr == val
    return isinstance(expr, _LiteralExpression) and expr.value == val


def to_expression(val: ExpressionLike) -> Expression:
    if isinstance(val, Expression):
        return val
    if isinstance(val, (float, int)):
        return literal(val)
    raise TypeError(f"Unexpected expression: {val}")


@dataclasses.dataclass(eq=False, frozen=True)
class ExpressionReference(Expression):
    identifier: Identifier
    subscripts: tuple[Expression, ...]

    def render(self, _precedence=0) -> str:
        return render_identifier(self.identifier, *self.subscripts)


@dataclasses.dataclass(eq=False, frozen=True)
class _UnaryExpression(Expression):
    operator: str
    expression: Expression

    def render(self, _precedence=0) -> str:
        op = self.operator
        if op == "abs":
            return f"\\lvert {self.expression.render()} \\rvert"
        return f"\\left\\l{op} {self.expression.render()} \\right\\r{op}"


_binary_operator_precedences = {
    "mul": (4, 4, 4),
    "add": (1, 1, 1),
    "mod": (3, 3, 3),
    "sub": (1, 2, 1),
    "div": (0, 0, 5),
    "pow": (0, 0, 5),
}


@dataclasses.dataclass(eq=False, frozen=True)
class _BinaryExpression(Expression):
    operator: str
    left_expression: Expression
    right_expression: Expression

    def render(self, precedence=0) -> str:
        op = self.operator
        left_inner, right_inner, outer = _binary_operator_precedences[op]
        left = self.left_expression.render(left_inner)
        right = self.right_expression.render(right_inner)
        if op == "mul":
            if is_literal(self.left_expression, -1):
                rendered = f"{{-{right}}}"
            elif is_literal(self.right_expression, -1):
                rendered = f"{{-{left}}}"
            else:
                rendered = f"{left} {right}"
        elif op == "add":
            rendered = f"{left} + {right}"
        elif op == "mod":
            rendered = f"{left} \\bmod {right}"
        elif op == "sub":
            rendered = f"{left} - {right}"
        elif op == "div":
            rendered = f"\\frac{{{left}}}{{{right}}}"
        elif op == "pow":
            rendered = f"\\left({left}\\right)^{{{right}}}"
        else:
            raise Exception(f"Unexpected operator: {op}")
        if outer < precedence:
            rendered = f"\\left({rendered}\\right)"
        return rendered


@dataclasses.dataclass(frozen=True)
class Domain:
    quantifiers: tuple[QuantifierIdentifier, ...]
    mask: Optional[Predicate] = None

    def render(self) -> str:
        groups = []
        outer = itertools.groupby(self.quantifiers, _quantifier_grouping_key)
        for (_id, key), outer_qs in outer:
            if isinstance(key, ScalarSpace):
                names = ", ".join(q.format() for q in outer_qs)
                groups.append(f"{names} \\in {key.render()}")
            else:
                inner = itertools.groupby(outer_qs, lambda q: q.outer_group)
                components: list[str] = []
                for g, inner_qs in inner:
                    group_names = list(q.format() for q in inner_qs)
                    joined = ", ".join(group_names)
                    components.append(
                        f"({joined})" if len(group_names) > 1 else joined
                    )
                group = ", ".join(components)
                group += " \\in "
                group += render_identifier(key.alias, *key.subscripts)
                groups.append(group)
        rendered = ", ".join(groups)
        if self.mask is not None:
            rendered += f" \\mid {self.mask.render()}"
        return rendered


def _quantifier_grouping_key(
    q: QuantifierIdentifier,
) -> tuple[int, Union[ScalarSpace, QuantifierGroup]]:
    # We add the ID to prevent `__eq__` from being called on equations
    sp = q.space
    if not isinstance(sp, ScalarSpace):
        raise TypeError(f"Unexpected space: {sp}")
    if not q.groups:
        return (id(sp), sp)
    g = q.groups[0]
    return (id(g), g)


@dataclasses.dataclass(eq=False, frozen=True)
class _SummationExpression(Expression):
    summand: Expression
    domain: Domain

    def render(self, precedence=0) -> str:
        inner = max(3, precedence)
        with local_formatting_scope(self.domain.quantifiers):
            rendered = f"\\sum_{{{self.domain.render()}}} "
            rendered += self.summand.render(inner)
        return rendered


@dataclasses.dataclass(eq=False, frozen=True)
class _CardinalityExpression(Expression):
    domain: Domain

    def render(self, _precedence=0) -> str:
        qs = self.domain.quantifiers
        with local_formatting_scope(qs):
            if len(qs) == 1 and self.domain.mask is None:
                _id, key = _quantifier_grouping_key(qs[0])
                if isinstance(key, ScalarSpace):
                    sp = key.render()
                else:
                    sp = render_identifier(key.alias, *key.subscripts)
            else:
                sp = f"\\{{ {self.domain.render()} \\}}"
            return f"\\# {sp}"


@dataclasses.dataclass(frozen=True)
class _SwitchCase:
    expression: Expression
    predicate: Optional[Predicate] = None


@dataclasses.dataclass(eq=False, frozen=True)
class _SwitchExpression(Expression):
    cases: Sequence[_SwitchCase]

    def render(self, precedence=0) -> str:
        cs: list[str] = []
        for c in self.cases:
            s = c.expression.render()
            if c.predicate is not None:
                s += f" \\mid {c.predicate.render()}"
            cs.append(s)
        sep = ", \\\\ "
        return f"\\begin{{cases}} {sep.join(cs)} \\end{{cases}}"


class Space:
    """Base quantification


    This class provides support for generating cross-products with the `*`
    operator (see :func:`~opvious.modeling.cross`):

    .. code-block:: python

        space1 * space2 # Equivalent to cross(space1, space2)
    """

    def __mul__(self, other: Quantifiable) -> Quantification:
        return cross(self, other)

    def __rmul__(self, left: Quantifiable) -> Quantification:
        return cross(left, self)


class ScalarSpace(Space):
    def __iter__(self) -> Quantified[Quantifier]:
        return (untuple(t) for t in cross(self))

    def render(self) -> str:
        raise NotImplementedError()


@dataclasses.dataclass(frozen=True)
class QuantifiableReference(ScalarSpace):
    identifier: AliasIdentifier
    subscripts: tuple[Expression, ...]
    quantifiers: tuple[QuantifierIdentifier, ...]

    def render(self) -> str:
        return render_identifier(self.identifier, *self.subscripts)


@dataclasses.dataclass(frozen=True)
class Quantifier(Expression):
    """An expression used to index a quantifiable space

    Quantifiers are generated by using :func:`cross` and its convenience
    alternatives (for example iterating on a :class:`Dimension`). You should
    not need to instantiate them directly - this class is only exposed for
    typing purposes.
    """

    identifier: QuantifierIdentifier

    def render(self, _precedence=0) -> str:
        return self.identifier.format()


_Q = TypeVar(
    "_Q", bound=Union[Quantifier, Sequence[Quantifier]], covariant=True
)


class IterableSpace(Protocol[_Q]):
    """Base protocol for spaces which can also be directly iterated on

    It is exposed mostly as a typing convenience for typing model fragments.
    :class:`~opvious.modeling.Space` is typically used for providing the
    underlying implementation.
    """

    def __mul__(self, other: Quantifiable) -> Quantification:
        raise NotImplementedError()

    def __rmul__(self, other: Quantifiable) -> Quantification:
        raise NotImplementedError()

    def __iter__(self) -> Iterator[_Q]:
        raise NotImplementedError()


def expression_space(expr: Expression) -> Optional[ScalarSpace]:
    """Returns the underlying scalar quantifiable for an expression if any"""
    if isinstance(expr, Quantifier):
        return expr.identifier.space
    return None


class Predicate:
    """A predicate on expressions

    Instances of this class are generated by using comparison operators on
    :class:`.Expression` instances. The example below shows the three types of
    predicates supported as constraints:

    .. code-block:: python

        class ProductModel(Model):
            products = Dimension()
            cost = Parameter(products)
            count = Variable(products)

            @constraint
            def total_cost_is_at_most_100(self):
                total_cost = total(
                    self.cost(p) * self.count(p)
                    for p in self.products
                )
                yield total_cost <= 100  # LEQ predicate

            @constraint
            def each_count_is_at_least_10(self):
                for p in self.products:
                    yield self.count(p) >= 10  # GEQ predicate

            @constraint
            def exactly_10_of_expensive_items(self):
                for p in self.products:
                    if self.cost(p) > 50:  # See below
                        yield self.count(p) == 10  # EQ predicate

    Additional types of predicates may be generated on expressions which do not
    include variables and used as conditionals within a quantified expression.
    For example the `exactly_10_of_expensive_items` constraint above uses a
    greater than predicate to filter the set of products where the constraint
    applies.
    """

    def __and__(self, other):
        return _BinaryPredicate("\\land", self, other)

    def __or__(self, other):
        return _BinaryPredicate("\\lor", self, other)

    def __bool__(self):
        declare(self)
        return True

    def render(self, precedence=0) -> str:
        raise NotImplementedError()


@dataclasses.dataclass(frozen=True)
class _ComparisonPredicate(Predicate):
    command: str
    left_expression: Expression
    right_expression: Expression

    def render(self, _precedence=0) -> str:
        left = self.left_expression.render()
        right = self.right_expression.render()
        return f"{left} {self.command} {right}"


_binary_condition_precedences = {
    "and": 2,
    "or": 1,
}


@dataclasses.dataclass(frozen=True)
class _BinaryPredicate(Predicate):
    condition: str
    left_predicate: Predicate
    right_predicate: Predicate

    def render(self, precedence=0) -> str:
        cond = self.condition
        inner = _binary_condition_precedences[cond]
        left = self.left_predicate.render(inner)
        right = self.right_predicate.render(inner)
        rendered = f"{left} \\l{cond} {right}"
        if inner < precedence:
            rendered = f"\\left({rendered}\\right)"
        return rendered


Quantifiable = Union[
    Iterable[Union[Quantifier, Sequence[Quantifier]]],  # Includes quantified
    Space,
    Domain,
    tuple["Quantifiable", ...],
]


_V = TypeVar("_V")


def within_domain(quantified: Quantified[_V]) -> tuple[_V, Domain]:
    value, declarations = unquantify(quantified)
    quantifiers: list[QuantifierIdentifier] = []
    mask: Optional[Predicate] = None
    for declaration in declarations:
        if isinstance(declaration, Predicate):
            if mask is None:
                mask = declaration
            else:
                mask = _BinaryPredicate("and", mask, declaration)
        elif isinstance(declaration, QuantifierIdentifier):
            quantifiers.append(declaration)
        else:
            raise TypeError(f"Unexpected declaration: {declaration}")
    domain = Domain(tuple(quantifiers), mask)
    return value, domain


def domain(
    quantifiable: Quantifiable,
    names: Optional[Iterable[Name]] = None,
) -> Domain:
    """Creates a domain from a quantifiable"""
    return _domain_from_quantified(iter(cross(quantifiable, names=names)))


def _domain_from_quantified(
    quantified: Quantified[Union[Quantifier, Iterable[Quantifier]]]
) -> Domain:
    qs, domain = within_domain(quantified)
    if isinstance(qs, Quantifier):
        idens: tuple[QuantifierIdentifier, ...] = (qs.identifier,)
    else:
        idens = tuple(q.identifier for q in qs)
    if not _isomorphic(idens, domain.quantifiers):
        raise Exception(
            f"Inconsistent quantifiers: {idens} != {domain.quantifiers}"
        )
    return dataclasses.replace(domain, quantifiers=idens)


def _isomorphic(
    qs1: Iterable[QuantifierIdentifier], qs2: Iterable[QuantifierIdentifier]
) -> bool:
    return collections.Counter(qs1) == collections.Counter(qs2)


Projection = int


@dataclasses.dataclass(frozen=True)
class Cross(Sequence[Quantifier]):
    """Cross-product result"""

    _quantifiers: tuple[Quantifier, ...]
    _lifted: Optional[tuple[Quantifier, ...]]

    @property
    def lifted(self) -> tuple[Quantifier, ...]:
        if self._lifted is None:
            raise Exception("Unlifted cross-product")
        return self._lifted

    def __len__(self) -> int:
        return len(self._quantifiers)

    @overload
    def __getitem__(self, ix: int) -> Quantifier:
        ...

    @overload
    def __getitem__(self, sl: slice) -> Sequence[Quantifier]:
        ...

    def __getitem__(self, arg: Any) -> Any:
        return self._quantifiers[arg]

    def __iter__(self):
        return iter(self._quantifiers)


@dataclasses.dataclass(
    frozen=True,
    eq=False,
)
class Quantification(Space):
    """Cross-product quantification"""

    _quantifiables: tuple[Quantifiable, ...]
    _names: Mapping[int, Name]
    _projection: Projection
    _lift: bool

    __hash__: Any = None

    def __iter__(self) -> Quantified[Cross]:
        projected: list[Quantifier] = []
        lifted: list[Quantifier] = []
        for i, d in enumerate(self._quantifiables):
            project = (1 << i) & self._projection
            if not project and not self._lift:
                continue
            j0 = len(projected)
            quants = list(
                Quantifier(declare(iden.named(self._names.get(j0 + j))))
                for j, iden in enumerate(_quantifier_identifiers(d))
            )
            lifted.extend(quants)
            if project:
                projected.extend(quants)
        yield Cross(tuple(projected), tuple(lifted))


def lift(
    projected: tuple[Quantifier, ...],
    unprojected: tuple[Quantifier, ...],
    projection: Projection,
) -> tuple[Quantifier, ...]:
    """Combines quantifiers to reconstruct an underlying quantification"""
    quants: list[Quantifier] = []
    i = j = 0
    for k in range(len(projected) + len(unprojected)):
        if (1 << k) & projection:
            quants.append(projected[i])
            i += 1
        else:
            quants.append(unprojected[j])
            j += 1
    return tuple(quants)


def cross(
    *quantifiables: Quantifiable,
    names: Optional[Iterable[Name]] = None,
    projection: Projection = -1,
    lift=False,
) -> Quantification:
    """Generates the cross-product of multiple quantifiables

    Args:
        quantifiables: One or more quantifiables
        names: Optional names for the generated quantifiers
        projection: Quantifiable selection mask
        lift: Returns lifted :class:`~opvious.modeling.Cross` instances.
            Setting this option will include all masks present in the original
            quantifiable, even if they are not projected.

    This function is the core building block for quantifying values.
    """
    return Quantification(
        _quantifiables=quantifiables,
        _names=dict(enumerate(names or [])),
        _projection=projection,
        _lift=lift,
    )


def _quantifier_identifiers(
    quantifiable: Quantifiable,
) -> Iterable[QuantifierIdentifier]:
    if isinstance(quantifiable, tuple):
        for component in quantifiable:
            yield from _quantifier_identifiers(component)
    elif isinstance(quantifiable, QuantifiableReference):
        qs = quantifiable.quantifiers
        group = QuantifierGroup(
            alias=quantifiable.identifier,
            subscripts=quantifiable.subscripts,
            rank=len(qs),
        )
        for q in qs:
            yield q.grouped_within(group)
    elif isinstance(quantifiable, ScalarSpace):
        yield QuantifierIdentifier.base(quantifiable)
    else:  # domain or quantified
        if isinstance(quantifiable, Domain):
            domain = quantifiable
        else:
            domain = _domain_from_quantified(cast(Any, quantifiable))
        if domain.mask is not None:
            declare(domain.mask)
        yield from domain.quantifiers


def total(body: Quantified[Expression]) -> Expression:
    """Returns an expression representing the sum of the quantified input

    Args:
        body: A quantified expression to be summed over
    """
    return _SummationExpression(*within_domain(body))


def size(quantifiable: Quantifiable) -> Expression:
    """Returns the cardinality of the quantifiable as an expression"""
    return _CardinalityExpression(domain(quantifiable))


def switch(
    *cases: Union[tuple[Predicate, ExpressionLike], ExpressionLike]
) -> Expression:
    """Returns an expression allowing branching between different values

    Args:
        cases: Tuples of expressions and optionally, predicate. The expression
            for the first predicate that matches will be used. A missing
            predicate always matches.
    """
    cs: list[_SwitchCase] = []
    for t in cases:
        if isinstance(t, tuple):
            cs.append(_SwitchCase(to_expression(t[1]), t[0]))
        else:
            cs.append(_SwitchCase(to_expression(t)))
    return _SwitchExpression(cs)



from __future__ import annotations

import contextvars
import dataclasses
import itertools
from typing import Any, Iterator, Tuple, TypeVar


_V = TypeVar("_V")


Quantified = Iterator[_V]


def _run_quantified(quantified: Quantified[_V]) -> _V:
    elems = list(itertools.islice(quantified, 2))
    if not elems:
        raise ValueError("Empty quantified")
    if len(elems) > 1:
        raise ValueError("Quantified contained multiple values")
    return elems[0]


@dataclasses.dataclass
class _Scope:
    declarations: list[Any]


_active_scope: Any = contextvars.ContextVar("quantified_scope")


def unquantify(quantified: Quantified[_V]) -> Tuple[_V, list[Any]]:
    scope = _Scope([])
    token = _active_scope.set(scope)
    try:
        value = _run_quantified(quantified)
    finally:
        _active_scope.reset(token)
    return value, scope.declarations


def declare(declaration: Any) -> Any:
    scope = _active_scope.get()
    scope.declarations.append(declaration)
    return declaration
